import sys
from unittest import TestCase
from io import StringIO
from decimal import Decimal
import math
import operator

from rpnpy import Calculator
from rpnpy.modifiers import Modifiers, strToModifiers


class TestDebug(TestCase):
    "Test setting/toggling debug output"

    def testDebugOffByDefault(self):
        "debugging must be off by default"
        c = Calculator()
        self.assertFalse(c._debug)

    def testStartWithDebugOn(self):
        "debugging must be settable via __init__"
        c = Calculator(debug=True)
        self.assertTrue(c._debug)

    def testDebugOn(self):
        "It must be possible to turn debugging on"
        c = Calculator()
        c.toggleDebug(True)
        self.assertTrue(c._debug)

    def testDebugOff(self):
        "It must be possible to turn debugging off"
        c = Calculator()
        c.toggleDebug(False)
        self.assertFalse(c._debug)

    def testToggleDebugOn(self):
        "It must be possible to toggle debugging on"
        c = Calculator()
        # Double-check it starts out off (don't rely on above test).
        self.assertFalse(c._debug)
        c.toggleDebug()
        self.assertTrue(c._debug)

    def testToggleDebugOff(self):
        "It must be possible to toggle debugging on"
        c = Calculator()
        # Double-check it starts out off (don't rely on above test).
        c.toggleDebug()
        self.assertTrue(c._debug)
        c.toggleDebug()
        self.assertFalse(c._debug)

    def testDebugOnViaModifier(self):
        "It must be possible to turn debugging on using a modifier"
        c = Calculator()
        c.execute(':D')
        self.assertTrue(c._debug)

    def testDebugOffViaModifier(self):
        "It must be possible to turn debugging off using a modifier"
        c = Calculator(debug=True)
        c.execute(':D')
        self.assertFalse(c._debug)


class TestCalculator(TestCase):

    def testEmptyStack(self):
        "A calculator must start life with an empty stack"
        self.assertEqual([], Calculator().stack)

    def testPushNumber(self):
        "Pushing a number must work as expected"
        c = Calculator()
        c.execute('4')
        (result,) = c.stack
        self.assertEqual(4, result)

    def testPop(self):
        "Calling pop must work as expected"
        c = Calculator()
        c.execute('4 5 pop')
        (result,) = c.stack
        self.assertEqual(4, result)

    def testPop2(self):
        "Calling pop with a count of 2 must work as expected"
        c = Calculator()
        c.execute('3 4 5 pop:2')
        (result,) = c.stack
        self.assertEqual(3, result)

    def testPopAll(self):
        "Calling pop on the whole stack must work as expected"
        c = Calculator()
        c.execute('3 4 5 pop:*')
        self.assertEqual([], c.stack)

    def testPushString(self):
        "Pushing a string must work as expected"
        c = Calculator()
        c.execute("'4'")
        (result,) = c.stack
        self.assertEqual('4', result)

    def testPushAbs(self):
        "Pushing the operator.abs function must work as expected"
        c = Calculator()
        c.execute('operator.abs :!')
        (result,) = c.stack
        self.assertIs(operator.abs, result)

    def testAdd(self):
        "Adding must work"
        c = Calculator()
        c.execute('3 4 +')
        (result,) = c.stack
        self.assertEqual(7, result)

    def testAddOneArg(self):
        "Adding must give an error if there is only one thing on the stack"
        err = StringIO()
        c = Calculator(errfp=err)
        c.execute('3 +')
        self.assertEqual(
            'Not enough args on stack! (+ needs 2 args, stack has 1 item)\n',
            err.getvalue())

    def testRegisterWithArgCount(self):
        """Registering and calling a new function and passing its argument
           count must work"""
        def double(n):
            return 2 * n

        c = Calculator()
        c.register(double, name='double', nArgs=1)
        c.execute('3 double')
        (result,) = c.stack
        self.assertEqual(6, result)

    def testRegisterWithArgCount2(self):
        """Registering and calling a new function and passing its argument
           count must work"""
        def doubleMax(*args):
            return 2 * max(*args)

        c = Calculator()
        c.register(doubleMax, name='doubleMax', nArgs=3)
        c.execute('3 5 6 7 doubleMax')
        (onStack, result) = c.stack
        self.assertEqual(14, result)
        self.assertEqual(3, onStack)

    def testRegisterWithoutArgCount(self):
        """Registering and calling a new function without passing its argument
           count must work"""
        def addAndDouble(i, j):
            return 2 * (i + j)

        c = Calculator()
        c.register(addAndDouble, name='addAndDouble')
        c.execute('3 5 addAndDouble')
        (result,) = c.stack
        self.assertEqual(16, result)

    def testRegisterCelciusNoExplicitName(self):
        """Register and call a function to convert Fahrenheit to Celcius but do
        not pass a function name"""
        def celcius(f):
            return (f - 32) / 1.8

        c = Calculator()
        c.register(celcius)
        c.execute('212 celcius')
        (result,) = c.stack
        self.assertEqual(100, result)

    def testRegisterLambdaWithExplicitName(self):
        "Register and call a function defined using lambda with a passed name"
        c = Calculator()
        c.register(lambda: 3, name='xx')
        c.execute('xx')
        (result,) = c.stack
        self.assertEqual(3, result)

    def testRegisterLambdaWithNoName(self):
        "Register and call a function defined using lambda and no passed name"
        c = Calculator()
        c.register(lambda: 3)
        # The function will be called '<lambda>', which is a bit dangerous
        # (could easily be accidentally be overwritten).
        c.execute('<lambda>')
        (result,) = c.stack
        self.assertEqual(3, result)

    def testSumOneArg(self):
        "sum must take one arg by default"
        c = Calculator()
        self.assertEqual(1, c._functions['sum'].nArgs)

    def testPowTwoArgs(self):
        "pow must take two args by default"
        c = Calculator()
        self.assertEqual(2, c._functions['pow'].nArgs)

    def testSwap(self):
        "swap must work as expected"
        c = Calculator()
        c.execute('4 5')
        self.assertEqual([4, 5], c.stack)
        c.execute('swap')
        self.assertEqual([5, 4], c.stack)

    def testDup(self):
        "dup must work as expected"
        c = Calculator()
        c.execute('4 d')
        self.assertEqual([4, 4], c.stack)
        c.execute('dup')
        self.assertEqual([4, 4, 4], c.stack)

    def testIterateString(self):
        "The :i (iterate) modifier must work as expected on a string"
        c = Calculator()
        c.execute('"hello" :i')
        (result,) = c.stack
        self.assertEqual(['h', 'e', 'l', 'l', 'o'], result)

    def testIterateGenerator(self):
        "The :i (iterate) modifier must work as expected on a generator"
        c = Calculator()
        c.execute('str :!')
        c.execute('[1,2,3]')
        c.execute('map :i')
        (result,) = c.stack
        self.assertEqual(['1', '2', '3'], result)

    def testItemgetter(self):
        "itemgetter must work correctly"
        c = Calculator(splitLines=False)
        c.execute('1')
        c.execute('itemgetter')
        c.execute('[[1, 2, 3], [4, 5, 6]]')
        c.execute('map :i')
        (result,) = c.stack
        self.assertEqual([2, 5], result)

    def testList(self):
        "Converting the top of the stack to a list must work"
        c = Calculator()
        c.execute('"hey" list')
        (result,) = c.stack
        self.assertEqual(['h', 'e', 'y'], result)

    def testList2(self):
        "Converting the top two items of the stack to a list must work"
        c = Calculator()
        c.execute('4 5 list:2')
        c.printStack()
        (result,) = c.stack
        self.assertEqual([4, 5], result)

    def testListAll(self):
        "Converting all items of the stack to a list must work"
        c = Calculator()
        c.execute('4 5 6 list:*')
        c.printStack()
        (result,) = c.stack
        self.assertEqual([4, 5, 6], result)

    def testPushList(self):
        "It must be possible to push list onto the stack"
        c = Calculator()
        c.execute('list :!')
        (result,) = c.stack
        self.assertIs(list, result)

    def testPushNone(self):
        "It must be possible to push None onto the stack"
        c = Calculator()
        c.execute('None')
        (result,) = c.stack
        self.assertIs(None, result)

    def testPushTrue(self):
        "It must be possible to push True onto the stack"
        c = Calculator()
        c.execute('True')
        (result,) = c.stack
        self.assertIs(True, result)

    def testPushFalse(self):
        "It must be possible to push False onto the stack"
        c = Calculator()
        c.execute('False')
        (result,) = c.stack
        self.assertIs(False, result)

    def testBatch(self):
        "It must be possible to read from a file using the batch method"
        infp = StringIO('3\n4\n+\np\n')
        outfp = StringIO()
        c = Calculator(outfp=outfp)
        c.batch(infp)
        self.assertEqual('7\n', outfp.getvalue())

    def testREPL(self):
        "It must be possible to read input in REPL form"
        infp = StringIO('3\n4\n+\np\n')
        outfp = StringIO()
        c = Calculator(outfp=outfp)
        stdin = sys.stdin
        sys.stdin = infp
        c.repl('prompt')
        sys.stdin = stdin
        self.assertEqual('7\n', outfp.getvalue())

    def testCountConflictsWithAllModifier(self):
        "If :* and a count is given, it's an error if they don't agree"
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        self.assertFalse(c.execute('4 5 6 list:*5'))
        error = ('* modifier conflicts with explicit count 5 '
                 '(stack has 3 items)\n')
        self.assertEqual(error, errfp.getvalue())

    def testFloat(self):
        "It must be possible to convert something to a float"
        c = Calculator()
        c.execute('4 float')
        (result,) = c.stack
        self.assertEqual(4.0, result)


class TestReverseModifier(TestCase):
    "Test the reverse modifier"

    def testReverseSubtractionArgs(self):
        "Reverse the args in a subtraction"
        c = Calculator()
        c.execute('5 4 -:r')
        (result,) = c.stack
        self.assertEqual(-1, result)

    def testMap(self):
        "Call map with args reversed from their normal order"
        c = Calculator()
        c.execute('[1,2,3]')
        c.execute('str :!')
        c.execute('map :ir')
        (result,) = c.stack
        self.assertEqual(['1', '2', '3'], result)


class TestReverseSpecialCommand(TestCase):
    "Test the reverse special command"

    def testReverse1(self):
        "Reversing just the top element of the stack does nothing"
        c = Calculator()
        c.execute('4 5 reverse:1')
        self.assertEqual([4, 5], c.stack)

    def testReverse(self):
        "Reverse (by default) the top two elements of the stack"
        c = Calculator()
        c.execute('4 5 reverse')
        self.assertEqual([5, 4], c.stack)

    def testReverseAll(self):
        "It must be possible to reverse all the stack"
        c = Calculator()
        c.execute('4 6 5 reverse:*')
        self.assertEqual([5, 6, 4], c.stack)

    def testReverseWithStackTooSmall(self):
        "It's an error if an attempt is made to reverse too many things"
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        self.assertFalse(c.execute('4 5 6 reverse:10'))
        error = 'Cannot reverse 10 items (stack length is 3)\n'
        self.assertEqual(error, errfp.getvalue())


class TestDecimal(TestCase):
    "Test working with Decimal instances"

    def testPush(self):
        "It must be possible to push Decimal by itself"
        c = Calculator()
        c.execute('Decimal :!')
        (result,) = c.stack
        self.assertIs(Decimal, result)

    def testPushValue(self):
        "It must be possible to push a Decimal value"
        c = Calculator()
        c.execute('Decimal(4)')
        (result,) = c.stack
        self.assertEqual(Decimal(4), result)

    def testArithmetic(self):
        "It must be possible to operate on Decimals"
        c = Calculator()
        c.execute('Decimal(4) Decimal(6) +')
        (result,) = c.stack
        self.assertEqual(Decimal(10), result)
        c.execute('int')
        (result,) = c.stack
        self.assertEqual(10, result)


class TestJoin(TestCase):
    "Test the join special function"

    def testEmptyString(self):
        "Joining on an empty string must work"
        c = Calculator()
        c.execute('"" ["4","5","6"] join')
        (result,) = c.stack
        self.assertEqual('456', result)

    def testNonEmptyString(self):
        "Joining on a non-empty string must work"
        c = Calculator()
        c.execute('"-" ["4","5","6"] join')
        (result,) = c.stack
        self.assertEqual('4-5-6', result)

    def testNonStrings(self):
        "Joining things that are not string must work"
        c = Calculator()
        c.execute('"-" [4,5,6] join')
        (result,) = c.stack
        self.assertEqual('4-5-6', result)

    def testWithCount(self):
        "Joining several stack items must work"
        c = Calculator()
        c.execute('3 "-" 4 5 6 join:3')
        self.assertEqual([3, '4-5-6'], c.stack)

    def testAllStack(self):
        "Joining the whole stack"
        c = Calculator()
        c.execute('"-" 3 4 5 6 join:*')
        (result,) = c.stack
        self.assertEqual('3-4-5-6', result)

    def testEmptyStringReversed(self):
        "Joining on an empty string must work"
        c = Calculator()
        c.execute('["4","5","6"] "" join:r')
        (result,) = c.stack
        self.assertEqual('456', result)

    def testNonEmptyStringReversed(self):
        "Joining on a non-empty string must work"
        c = Calculator()
        c.execute('["4","5","6"] "-" join:r')
        (result,) = c.stack
        self.assertEqual('4-5-6', result)

    def testNonStringsReversed(self):
        "Joining things that are not string must work"
        c = Calculator()
        c.execute('[4,5,6] "-" join:r')
        (result,) = c.stack
        self.assertEqual('4-5-6', result)

    def testWithCountReversed(self):
        "Joining several stack items must work"
        c = Calculator()
        c.execute('3 4 5 6 "-" join:3r')
        self.assertEqual([3, '4-5-6'], c.stack)

    def testAllStackReversed(self):
        "Joining the whole stack"
        c = Calculator()
        c.execute('3 4 5 6 "-" join:*r')
        (result,) = c.stack
        self.assertEqual('3-4-5-6', result)


class TestFindCallableAndArgs(TestCase):
    "Test the findCallableAndArgs function"

    def testEmptyStack(self):
        "Calling on an empty stack must return None, None"
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        func, args = c.findCallableAndArgs('cmd', Modifiers(), None)
        self.assertEqual((None, None), (func, args))
        error = "Cannot run 'cmd' (stack has only 0 items)\n"
        self.assertEqual(error, errfp.getvalue())

    def testStackLengthOne(self):
        "Calling on a stack with only one item must return None, None"
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        c.execute('4')
        func, args = c.findCallableAndArgs('cmd', Modifiers(), None)
        self.assertEqual((None, None), (func, args))
        error = "Cannot run 'cmd' (stack has only 1 item)\n"
        self.assertEqual(error, errfp.getvalue())

    def testStackLengthTwoNoCount(self):
        """Calling on a stack with two items and no count must return
        correctly. The number of stack items is obtained from the function
        signature"""
        c = Calculator()
        c.execute('log10 :!')
        c.execute('4')
        func, args = c.findCallableAndArgs('cmd', Modifiers(), None)
        self.assertIs(math.log10, func)
        self.assertEqual([4], args)

    def testStackLengthTwoWithCount(self):
        """Calling on a stack with two items and a count must return
        correctly."""
        c = Calculator()
        c.execute('log10 :!')
        c.execute('4')
        func, args = c.findCallableAndArgs('cmd', Modifiers(), 1)
        self.assertIs(math.log10, func)
        self.assertEqual([4], args)

    def testStackLengthThreeWithCountError(self):
        """Calling on a stack with three items and a count that points to a
        non-callable must result in an error"""
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        c.execute('log10 :!')
        c.execute('4')
        c.execute('5')
        func, args = c.findCallableAndArgs('cmd', Modifiers(), 1)
        self.assertIs(None, func)
        self.assertEqual(None, args)
        error = ("Cannot run 'cmd' with 1 argument. Stack item (4) is "
                 "not callable\n")
        self.assertEqual(error, errfp.getvalue())

    def testStackLengthThree(self):
        "Calling on a stack with three items (count=2) must return correctly"
        c = Calculator()
        c.execute('log10 :!')
        c.execute('4')
        c.execute('5')
        func, args = c.findCallableAndArgs('cmd', Modifiers(), 2)
        self.assertIs(math.log10, func)
        self.assertEqual([4, 5], args)

    def testAllStack(self):
        "Calling on all the stack must return correctly"
        c = Calculator()
        c.execute('log10 :!')
        c.execute('4')
        c.execute('5')
        c.execute('6')
        func, args = c.findCallableAndArgs('cmd', strToModifiers('*'), None)
        self.assertIs(math.log10, func)
        self.assertEqual([4, 5, 6], args)


class TestFindCallableAndArgsReversed(TestCase):
    "Test the findCallableAndArgs function when the reversed modifier is used"

    def testEmptyStack(self):
        "Calling on an empty stack must return None, None"
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        func, args = c.findCallableAndArgs('cmd', strToModifiers('r'), None)
        self.assertEqual((None, None), (func, args))
        error = "Cannot run 'cmd' (stack has only 0 items)\n"
        self.assertEqual(error, errfp.getvalue())

    def testStackLengthOne(self):
        "Calling on a stack with only one item must return None, None"
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        c.execute('4')
        func, args = c.findCallableAndArgs('cmd', strToModifiers('r'), None)
        self.assertEqual((None, None), (func, args))
        error = "Cannot run 'cmd' (stack has only 1 item)\n"
        self.assertEqual(error, errfp.getvalue())

    def testStackLengthTwoNoCount(self):
        """Calling on a stack with two items and no count must return
        correctly. The number of stack items is obtained from the function
        signature"""
        c = Calculator()
        c.execute('4')
        c.execute('log10 :!')
        func, args = c.findCallableAndArgs('cmd', strToModifiers('r'), None)
        self.assertIs(math.log10, func)
        self.assertEqual([4], args)

    def testStackLengthTwoWithCount(self):
        """Calling on a stack with two items and a count must return
        correctly."""
        c = Calculator()
        c.execute('4')
        c.execute('log10 :!')
        func, args = c.findCallableAndArgs('cmd', strToModifiers('r'), 1)
        self.assertIs(math.log10, func)
        self.assertEqual([4], args)

    def testStackLengthThree(self):
        "Calling on a stack with three items (count=2) must return correctly"
        c = Calculator()
        c.execute('5')
        c.execute('4')
        c.execute('log10 :!')
        func, args = c.findCallableAndArgs('cmd', strToModifiers('r'), 2)
        self.assertIs(math.log10, func)
        self.assertEqual([5, 4], args)

    def testAllStack(self):
        "Calling on all the stack must return correctly"
        c = Calculator()
        c.execute('4')
        c.execute('5')
        c.execute('6')
        c.execute('log10 :!')
        func, args = c.findCallableAndArgs('cmd', strToModifiers('*r'), None)
        self.assertIs(math.log10, func)
        self.assertEqual([4, 5, 6], args)


class TestFindStringAndArgs(TestCase):
    "Test the findStringAndArgs function"

    def testEmptyStack(self):
        "Calling on an empty stack must return None, None"
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        string, args = c.findStringAndArgs('cmd', Modifiers(), None)
        self.assertEqual((None, None), (string, args))
        error = "Cannot run 'cmd' (stack has only 0 items)\n"
        self.assertEqual(error, errfp.getvalue())

    def testStackLengthOne(self):
        "Calling on a stack with only one item must return None, None"
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        c.execute('4')
        string, args = c.findStringAndArgs('cmd', Modifiers(), None)
        self.assertEqual((None, None), (string, args))
        error = "Cannot run 'cmd' (stack has only 1 item)\n"
        self.assertEqual(error, errfp.getvalue())

    def testStackLengthTwoNoCount(self):
        """Calling on a stack with two items and no count must return
        correctly. The number of stack items is obtained from the function
        signature"""
        c = Calculator()
        c.execute("'string'")
        c.execute('4')
        string, args = c.findStringAndArgs('cmd', Modifiers(), None)
        self.assertEqual('string', string)
        self.assertEqual([4], args)

    def testStackLengthTwoWithCount(self):
        """Calling on a stack with two items and a count must return
        correctly."""
        c = Calculator()
        c.execute("'string'")
        c.execute('4')
        string, args = c.findStringAndArgs('cmd', Modifiers(), 1)
        self.assertEqual('string', string)
        self.assertEqual([4], args)

    def testStackLengthThreeWithCountError(self):
        """Calling on a stack with three items and a count that points to a
        non-string must result in an error"""
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        c.execute("'string'")
        c.execute('4')
        c.execute('5')
        string, args = c.findStringAndArgs('cmd', Modifiers(), 1)
        self.assertIs(None, string)
        self.assertEqual(None, args)
        error = ("Cannot run 'cmd' with 1 argument. Stack item (4) is "
                 "not a string\n")
        self.assertEqual(error, errfp.getvalue())

    def testStackLengthThree(self):
        "Calling on a stack with three items (count=2) must return correctly"
        c = Calculator()
        c.execute("'string'")
        c.execute('4')
        c.execute('5')
        string, args = c.findStringAndArgs('cmd', Modifiers(), 2)
        self.assertEqual('string', string)
        self.assertEqual([4, 5], args)

    def testAllStack(self):
        "Calling on all the stack must return correctly"
        c = Calculator()
        c.execute("'string'")
        c.execute('4')
        c.execute('5')
        c.execute('6')
        string, args = c.findStringAndArgs('cmd', strToModifiers('*'), None)
        self.assertEqual('string', string)
        self.assertEqual([4, 5, 6], args)


class TestFindStringAndArgsReversed(TestCase):
    "Test the findStringAndArgs function when the reversed modifier is used"

    def testEmptyStack(self):
        "Calling on an empty stack must return None, None"
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        string, args = c.findStringAndArgs('cmd', strToModifiers('r'), None)
        self.assertEqual((None, None), (string, args))
        error = "Cannot run 'cmd' (stack has only 0 items)\n"
        self.assertEqual(error, errfp.getvalue())

    def testStackLengthOne(self):
        "Calling on a stack with only one item must return None, None"
        errfp = StringIO()
        c = Calculator(errfp=errfp)
        c.execute('4')
        string, args = c.findStringAndArgs('cmd', strToModifiers('r'), None)
        self.assertEqual((None, None), (string, args))
        error = "Cannot run 'cmd' (stack has only 1 item)\n"
        self.assertEqual(error, errfp.getvalue())

    def testStackLengthTwoNoCount(self):
        """Calling on a stack with two items and no count must return
        correctly. The number of stack items is obtained from the function
        signature"""
        c = Calculator()
        c.execute('4')
        c.execute("'string'")
        string, args = c.findStringAndArgs('cmd', strToModifiers('r'), None)
        self.assertEqual('string', string)
        self.assertEqual([4], args)

    def testStackLengthTwoWithCount(self):
        """Calling on a stack with two items and a count must return
        correctly."""
        c = Calculator()
        c.execute('4')
        c.execute("'string'")
        string, args = c.findStringAndArgs('cmd', strToModifiers('r'), 1)
        self.assertEqual('string', string)
        self.assertEqual([4], args)

    def testStackLengthThree(self):
        "Calling on a stack with three items (count=2) must return correctly"
        c = Calculator()
        c.execute('5')
        c.execute('4')
        c.execute("'string'")
        string, args = c.findStringAndArgs('cmd', strToModifiers('r'), 2)
        self.assertEqual('string', string)
        self.assertEqual([5, 4], args)

    def testAllStack(self):
        "Calling on all the stack must return correctly"
        c = Calculator()
        c.execute('4')
        c.execute('5')
        c.execute('6')
        c.execute("'string'")
        string, args = c.findStringAndArgs('cmd', strToModifiers('*r'), None)
        self.assertEqual('string', string)
        self.assertEqual([4, 5, 6], args)


class TestApply(TestCase):
    "Test the apply special function"

    def testApplyNoCount(self):
        "apply must work correctly when not given a count"
        c = Calculator()
        c.execute('abs :!')
        c.execute('-1')
        c.execute('apply')
        (result,) = c.stack
        self.assertEqual(1, result)

    def testApplyWithCount(self):
        "apply must work correctly when given a count"
        c = Calculator()
        c.execute('+ :!')
        c.execute('3 5')
        c.execute('apply :2')
        (result,) = c.stack
        self.assertEqual(8, result)


class TestReduce(TestCase):
    "Test the reduce special function"

    def testReduceAl(self):
        "Reduce must work correctly when told to operate on the whole stack"
        c = Calculator()
        c.execute('+ :!')
        c.execute('5')
        c.execute('6')
        c.execute('7')
        c.execute('reduce :*')
        (result,) = c.stack
        self.assertEqual(18, result)

    def testReduceWithCount(self):
        "Reduce must work correctly when given a count"
        c = Calculator()
        c.execute('+ :!')
        c.execute('5')
        c.execute('6')
        c.execute('reduce :2')
        (result,) = c.stack
        self.assertEqual(11, result)


class TestReduceReversed(TestCase):
    "Test the reduce special function when the reverse modifier is used"

    def testReduceAl(self):
        "Reduce must work correctly when told to operate on the whole stack"
        c = Calculator()
        c.execute('5')
        c.execute('6')
        c.execute('7')
        c.execute('+ :!')
        c.execute('reduce :*r')
        (result,) = c.stack
        self.assertEqual(18, result)

    def testReduceWithCount(self):
        "Reduce must work correctly when given a count"
        c = Calculator()
        c.execute('4')
        c.execute('5')
        c.execute('6')
        c.execute('+ :!')
        c.execute('reduce :2r')
        self.assertEqual([4, 11], c.stack)



from unittest import TestCase
from math import log

from rpnpy.inspect import countArgs


class TestCountArgs(TestCase):
    """Test the countArgs function"""

    def testZero(self):
        "A function that takes zero arguments must be processed correctly"
        self.assertEqual(0, countArgs(lambda: 3))

    def testOne(self):
        "A function that takes one argument must be processed correctly"
        self.assertEqual(1, countArgs(lambda x: 3))

    def testTwo(self):
        "A function that takes two arguments must be processed correctly"
        self.assertEqual(2, countArgs(lambda x, y: 3))

    def testLog(self):
        "The signature of math.log can't be inspected (at least in Python 3.7)"
        self.assertEqual(None, countArgs(log))

    def testLogWithDefault(self):
        """The signature of math.log can't be inspected (at least in Python
           3.7). Pass a default value."""
        self.assertEqual(3, countArgs(log, 3))



#!/usr/bin/env python

from __future__ import print_function

import sys
import os
import readline
import argparse

from rpnpy import Calculator


def setupReadline():
    """Initialize the readline library and command history.

    @return: A C{bool} to indicate whether standard input is a terminal
        (and therefore interactive).
    """
    if not os.isatty(0):
        # Standard input is closed or is a pipe etc. So there's no user
        # typing at us, and so no point in setting up readline.
        return False

    # Readline code from https://docs.python.org/3.7/library/readline.html
    histfile = os.path.join(os.path.expanduser('~'), '.pycalc_history')

    try:
        readline.read_history_file(histfile)
        historyLen = readline.get_current_history_length()
    except FileNotFoundError:
        open(histfile, 'wb').close()
        historyLen = 0

    try:
        readline.append_history_file
    except AttributeError:
        # We won't be able to save readline history. This can happen on
        # Python 3.5 at least - not sure why.
        pass
    else:
        import atexit

        def saveHistory(prevHistoryLen, histfile):
            newHistoryLen = readline.get_current_history_length()
            readline.set_history_length(1000)
            readline.append_history_file(newHistoryLen - prevHistoryLen,
                                         histfile)

        atexit.register(saveHistory, historyLen, histfile)

    return True


def parseArgs():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            'An RPN calculator for Python. Reads commands from standard input '
            'and/or files or interactively with a read-eval-print loop and '
            '(optionally) writes the final stack to standard output.'))

    parser.add_argument(
        '--prompt', default='--> ',
        help='The prompt to print at the start of each line when interactive.')

    parser.add_argument(
        'files', nargs='*',
        help=('Files to read input from. If you use this option and you also '
              'want standard input to be read at some point, use "-" as a '
              'name in the list of file names.'))

    parser.add_argument(
        '--separator',
        help=('The character to use to split lines of standard input into '
              'separate commands (unless --noSplit is given).'))

    parser.add_argument(
        '--debug', action='store_true', default=False,
        help='Print verbose information about how commandsa are run.')

    parser.add_argument(
        '--version', action='store_true', default=False,
        help='Print the version number and exit.')

    parser.add_argument(
        '--noSplit', action='store_false', default=True, dest='splitLines',
        help=('If given, do not split lines read from standard input into '
              'separate commands, treat each line as an entire command.'))

    parser.add_argument(
        '--noPrint', action='store_false', default=True, dest='print',
        help=('If given, do not print the stack after processing all commands '
              'from standard input.'))

    parser.add_argument(
        '--stdin', action='store_true', default=False,
        help=('If the arguments on the command line are passed as input to '
              'the calculator, you can use this option to also read commands '
              'from standard input once the command line has been executed.'))

    return parser.parse_args()


if __name__ == '__main__':
    args = parseArgs()

    if args.version:
        from rpnpy import __version__
        print(__version__)
    else:
        calc = Calculator(splitLines=args.splitLines, separator=args.separator,
                          debug=args.debug)

        interactive = setupReadline()

        if args.files:
            if all(os.path.exists(f) or f == '-' for f in args.files):
                # All arguments are existing files (or are '-', for stdin).
                for filename in args.files:
                    if filename == '-':
                        calc.repl(args.prompt)
                    else:
                        with open(filename) as fp:
                            calc.batch(fp, False)
            else:
                # Execute the command line as a set of commands, following
                # great suggestion by David Pattinson.
                calc.batch((' '.join(args.files),), args.print)
                if args.stdin:
                    calc.repl(args.prompt)
        elif interactive:
            calc.repl(args.prompt)
        else:
            calc.batch(sys.stdin, args.print)



from unittest import TestCase

from rpnpy.errors import UnknownModifiersError, IncompatibleModifiersError
from rpnpy.modifiers import Modifiers, strToModifiers, MODIFIERS


class TestModifiers(TestCase):
    "Test the Modifiers namedtuple and functions."

    def testUnknownModifiers(self):
        "Unknown modifiers must be detected."
        error = r"^\('x', 'y', 'z'\)$"
        self.assertRaisesRegex(
            UnknownModifiersError, error, strToModifiers, 'yxz')

    def testIncompatibleModifiersPushAndPreserve(self):
        "Incompatible modifiers must be detected."
        error = r'= \(preserve stack\) makes no sense with ! \(push\)'
        self.assertRaisesRegex(
            IncompatibleModifiersError, error, strToModifiers, '!=')

    def testIncompatibleModifiersSplitNoSplit(self):
        "Incompatible modifiers split and noSplit must be detected."
        error = r'\(split lines\) makes no sense with n \(do not split lines\)'

        self.assertRaisesRegex(
            IncompatibleModifiersError, error, strToModifiers, 'ns')

    def testKnownModifiers(self):
        "Make sure only known modifiers are present."
        self.assertEqual(sorted('*=!cDinprs'), sorted(MODIFIERS))

    def testNamesDiffer(self):
        "Make sure all known modifiers have different names."
        self.assertEqual(len(MODIFIERS), len(set(MODIFIERS.values())))

    def testDefaultsFalse(self):
        "Test all attributes default to False."
        modifiers = Modifiers()
        for name in MODIFIERS.values():
            self.assertEqual(False, getattr(modifiers, name))

    def testAll(self):
        "Test the strToModifiers function sets all."
        modifiers = strToModifiers('*')
        self.assertTrue(modifiers.all)

    def testDebug(self):
        "Test the strToModifiers function sets debug."
        modifiers = strToModifiers('D')
        self.assertTrue(modifiers.debug)

    def testForceCommand(self):
        "Test the strToModifiers function sets forceCommand."
        modifiers = strToModifiers('c')
        self.assertTrue(modifiers.forceCommand)

    def testIterate(self):
        "Test the strToModifiers function sets iterate."
        modifiers = strToModifiers('i')
        self.assertTrue(modifiers.iterate)

    def testPresevereStack(self):
        "Test the strToModifiers function sets preserveStack."
        modifiers = strToModifiers('=')
        self.assertTrue(modifiers.preserveStack)

    def testNoSplit(self):
        "Test the strToModifiers function sets noSplit."
        modifiers = strToModifiers('n')
        self.assertTrue(modifiers.noSplit)

    def testPrint(self):
        "Test the strToModifiers function sets print."
        modifiers = strToModifiers('p')
        self.assertTrue(modifiers.print)

    def testPush(self):
        "Test the strToModifiers function sets push."
        modifiers = strToModifiers('!')
        self.assertTrue(modifiers.push)

    def testReverse(self):
        "Test the strToModifiers function sets reverse."
        modifiers = strToModifiers('r')
        self.assertTrue(modifiers.reverse)

    def testSplit(self):
        "Test the strToModifiers function sets split."
        modifiers = strToModifiers('s')
        self.assertTrue(modifiers.split)



from unittest import TestCase

from rpnpy.io import findCommands, findModifiers
from rpnpy.modifiers import Modifiers, strToModifiers, MODIFIERS


class TestFindModifiers(TestCase):
    "Test the findModifiers function."

    def testEmpty(self):
        "Test an empty string."
        self.assertEqual((-1, Modifiers(), None), findModifiers(''))

    def testSeparatorIsLastLetter(self):
        "Test a string that ends with the separator."
        self.assertEqual((0, Modifiers(), None), findModifiers(':'))

    def testStringWithNoModifierSeparator(self):
        "Test a string with no modifier separator."
        self.assertEqual((-1, Modifiers(), None), findModifiers('hello'))

    def testUnknownLetter(self):
        "Test a string with a non-modifier letter."
        # Before testing, make sure '?' is not actually a valid modifier.
        self.assertFalse('?' in MODIFIERS)
        self.assertEqual((-1, Modifiers(), None), findModifiers('4 :?'))

    def testDict(self):
        "A dict must not be confused for a string with modifiers."
        self.assertEqual((-1, Modifiers(), None),
                         findModifiers("{'a':6, 'b':10, 'c':15}"))

    def testDictWithModifiers(self):
        "A dict with modifiers must be processed correctly."
        self.assertEqual((24, strToModifiers('=p'), 17),
                         findModifiers("{'a':6, 'b':10, 'c':15} :=p17"))

    def testDuplicatedLetter(self):
        "A string with a duplicated modifier must be processed correctly."
        self.assertEqual((4, strToModifiers('=p'), None),
                         findModifiers("abs :=pp"))


class TestFindCommands(TestCase):
    "Test the findCommands function."

    def testEmpty(self):
        "Test an empty string produces no commands at all."
        self.assertEqual(tuple(), tuple(findCommands('')))

    def testComment(self):
        "Test a comment."
        self.assertEqual((('', Modifiers(), None),),
                         tuple(findCommands('# comment')))

    def testCommentPrecededByWhitespace(self):
        "Test a comment preceeded by whitespace."
        self.assertEqual((('', Modifiers(), None),),
                         tuple(findCommands('  # comment')))

    def testCommandWithNoModifiers(self):
        "Test a command with no modifiers"
        self.assertEqual((('hey', Modifiers(), None),),
                         tuple(findCommands('hey')))

    def testTwoCommands(self):
        "Test two commands on one line"
        self.assertEqual(
            (('hey', Modifiers(), None), ('you', Modifiers(), None)),
            tuple(findCommands('hey you')))

    def testTwoCommandsFirstWithModifiers(self):
        "Test two commands on one line, the 1st with modifiers (and no space)"
        self.assertEqual(
            (('hey', strToModifiers('='), None), ('you', Modifiers(), None)),
            tuple(findCommands('hey:= you')))

    def testTwoCommandsFirstWithModifiersAfterSpace(self):
        "Test two commands on one line, the 2nd with modifiers after a space"
        self.assertEqual(
            (('hey', strToModifiers('='), None), ('you', Modifiers(), None)),
            tuple(findCommands('hey := you')))

    def testTwoCommandsSecondWithModifiers(self):
        "Test two commands on one line, the 2nd with modifiers (and no space)"
        self.assertEqual(
            (('hey', Modifiers(), None), ('you', strToModifiers('='), None)),
            tuple(findCommands('hey you:=')))

    def testTwoCommandsSecondWithModifiersAfterSpace(self):
        "Test two commands on one line, the 2nd with modifiers after a space"
        self.assertEqual(
            (('hey', Modifiers(), None), ('you', strToModifiers('='), None)),
            tuple(findCommands('hey you :=')))

    def testThreeCommandsSecondWithModifiers(self):
        "Test three commands on one line, the 2nd with modifiers"
        self.assertEqual(
            (('hey', Modifiers(), None), ('you', strToModifiers('='), None),
             ('there', Modifiers(), None)),
            tuple(findCommands('hey you:= there')))

    def testNoSplitAssignment(self):
        "Test an assignment with splitting off"
        self.assertEqual((('a = 3', Modifiers(), None),),
                         tuple(findCommands('a = 3', splitLines=False)))

    def testTwoCommandsSplitLines(self):
        "Test two words on one line when splitLines is False"
        self.assertEqual(
            (('hey you', Modifiers(), None),),
            tuple(findCommands('hey you', splitLines=False)))

    def testCommandWithNoModifiersPrecededByWhitespace(self):
        "Test a command with no modifiers preceded by whitespace"
        self.assertEqual((('hey', Modifiers(), None),),
                         tuple(findCommands(' hey')))

    def testCommandWithNoModifiersFollowedByWhitespace(self):
        "Test a command with no modifiers followed by whitespace"
        self.assertEqual((('hey', Modifiers(), None),),
                         tuple(findCommands('hey ')))

    def testCommandWithSurroundingSpaceAndNoModifiers(self):
        "Test a command with surrounding space and no modifiers"
        self.assertEqual((('hey', Modifiers(), None),),
                         tuple(findCommands(' hey ')))

    def testCommandWithNoModifiersCaseUnchanged(self):
        "Test that command case is unmodified"
        self.assertEqual((('HeY', Modifiers(), None),),
                         tuple(findCommands('HeY')))

    def testCommandWithCount(self):
        "Test a command with a count"
        self.assertEqual((('hey', Modifiers(), 101),),
                         tuple(findCommands('hey :101')))

    def testCommandWithCountThenModifiers(self):
        "Test a command with a count then modifiers"
        self.assertEqual((('hey', strToModifiers('=p'), 101),),
                         tuple(findCommands('hey :101=p')))

    def testCommandWithModifiersThenCount(self):
        "Test a command with modifiers then a count"
        self.assertEqual((('hey', strToModifiers('=p'), 101),),
                         tuple(findCommands('hey :=p101')))

    def testCommandWithCountThenSpaceAndModifiers(self):
        "Test a command with a count then a space then modifiers with space"
        self.assertEqual(
            (('hey', strToModifiers('=p'), 101),),
            tuple(findCommands('hey :101 =p', splitLines=False)))

    def testCommandWithModifiersSurroundingCount(self):
        "Test a command with modifiers before and after a count"
        self.assertEqual(
            (('hey', strToModifiers('=p*'), 16),),
            tuple(findCommands('hey :=p16*')))

    def testCommandWithModifiersSurroundingCountNoSplit(self):
        "Test a command with modifiers before and after a count"
        self.assertEqual(
            (('hey', strToModifiers('=p*'), 16),),
            tuple(findCommands('hey :=p 16 *', splitLines=False)))

    def testDict(self):
        "A dict must not be confused for a string with modifiers."
        self.assertEqual((("{'a':6,'b':10,'c':15}", Modifiers(), None),),
                         tuple(findCommands("{'a':6,'b':10,'c':15}")))

    def testDictWithModifiers(self):
        "A dict with a count must be processed correctly."
        self.assertEqual((("{'a':6,'b':10,'c':15}", Modifiers(), 7),),
                         tuple(findCommands("{'a':6,'b':10,'c':15}:7")))



#!/usr/bin/env python

from setuptools import setup


# Modified from http://stackoverflow.com/questions/2058802/
# how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
def version():
    import os
    import re

    init = os.path.join('rpnpy', '__init__.py')
    with open(init) as fp:
        initData = fp.read()
    match = re.search(r"^__version__ = ['\"]([^'\"]+)['\"]",
                      initData, re.M)
    if match:
        return match.group(1)
    else:
        raise RuntimeError('Unable to find version string in %r.' % init)


setup(name='rpnpy',
      version=version(),
      include_package_data=False,
      url='https://github.com/terrycojones/rpnpy',
      download_url='https://github.com/terrycojones/rpnpy',
      author='Terry Jones',
      author_email='tcj25@cam.ac.uk',
      keywords=['python reverse polish calculator'],
      classifiers=[
          'Programming Language :: Python :: 3',
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      license='MIT',
      scripts=['rpn.py'],
      description=('Control an RPN calculator from Python.'),
      extras_require={
        'dev': [
            'flake8',
            'pytest',
        ]
      })
